
8 constant max-testers
32 constant max-results  \ default for saved stack sizes

( string handling )

: allot-chars ( c-addr u -- c-addr' u )
	here swap 2dup 2>r chars dup allot move 2r> ;


( progress reporting )

variable verbose  false verbose !
: TESTING ( "...<cr>" -- )
	verbose @ if source type cr then
	postpone \ ;


( result sets )

\ There are two result sets; the expected (correct) results, and
\ the received (actual) results produced by the test code.  We
\ use a flag to indicate which one we are working with.

true constant expected
false constant received


( current tester )

variable this-tester
: (tvar) ( u.offset -- addr )  this-tester @ + ;
: tvar ( u.offset u.size "name" -- u.offset' )
	over >r : r> postpone literal postpone (tvar)
	postpone ;  + ;
0
	2 cells tvar tester-id      \ stack-name: c-addr u
	1 cells tvar begin-test-xt  ( -- )
	1 cells tvar save-stack-xt  ( ... flag.which -- )
	1 cells tvar #results-xt    ( flag.which -- n )
	0       tvar test-types     \ 2 * max-results xts
		\ result= ( u.index -- flag u.step )
		\ .result ( u.index flag.which -- u.step )
constant /tester

: .stack-id ( -- )  tester-id 2@ tuck type if space then ;
: result= ( u -- flag u2 )
	dup 2* cells test-types + perform ;
: .result ( u flag.which -- u2 )
	over 2* 1+ cells test-types + perform ;
: #results ( flag.which -- n )  #results-xt perform ;


( active testers )

variable #testers  0 #testers !
create testers max-testers cells allot
: push-tester ( addr.tester -- )
	dup #testers @ cells testers + !  1 #testers +!
	this-tester ! ;
: tester-bounds ( -- addr1 addr2 )
	testers #testers @ cells over + swap ;
: set-tester ( addr -- )  @ this-tester ! ;
: foreach-tester ( -- do-sys )
	postpone tester-bounds postpone ?do
		postpone i postpone set-tester ; immediate
: next-tester ( do-sys -- )
	1 cells postpone literal postpone +loop ; immediate


( defining new testers )

: new-tester ( c-addr u u.max-results -- )
	align here swap 2* cells /tester + allot push-tester
	allot-chars tester-id 2! ;
: test-type! ( xt.test xt.print addr u -- )
	2* cells +  tuck cell+ ! ! ;
\ convenience word for use in begin-test definitions
: default-types ( xt.test xt.print u -- )
	test-types swap 2* cells  over + swap do
		2dup i tuck cell+ ! !
	2 cells +loop 2drop ;
: #results: ( "name" -- )
	align here 2 cells allot >r
	>in @ >r : r> >in ! ( flag.which -- addr )
		1 cells postpone literal postpone and
		r> postpone literal postpone + postpone ;
	' >r :noname r> compile, postpone @ postpone ;
	#results-xt ! ;


( looping over results and testers )

: stack= ( -- flag )
	true 0 begin dup expected #results u< while nip
		dup result= rot +
	over 0= until then drop
	expected #results received #results = and ;
: results= ( -- flag )
	true foreach-tester stack= and next-tester ;
: .stack ( flag.which -- )
	." ( " .stack-id  dup #results 0< if
		#results . ." results "
	else
		0 swap begin 2dup #results u< while
			2dup .result rot + swap
		repeat 2drop
	then  ." ) " ;
: .results ( flag.which -- )
	foreach-tester dup .stack next-tester drop ;

: begin-results ( -- )
	foreach-tester begin-test-xt perform next-tester ;
: save-results ( ... flag.which -- )
	foreach-tester
		dup >r save-stack-xt perform r>
	next-tester drop ;


( failure flag )

variable should-fail  false should-fail !
: failure? ( -- flag )
	results= 0= should-fail @ <>  false should-fail ! ;
: test-failed ( -- )
	." Test Failed: " source type cr
	."   Expected: " expected .results cr
	."   Received: " received .results cr ;


( tester syntax )

: FAIL true should-fail ! ;
: T{ begin-results ;
: => received save-results ;  : -> => ;
: }T expected save-results
	failure? if test-failed then ;
